/* Copyright (C) 2006 Brian Aker, TangentOrg

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/*
  The memcache engine allows you to expose simple functions as tables
  to the storage engine interface.

  Happy coding!
    -Brian
*/

#ifdef USE_PRAGMA_IMPLEMENTATION
#pragma implementation        // gcc: Class implementation
#endif

#define MYSQL_SERVER 1
#include "mysql_priv.h"
#include "ha_memcache.h"
#include <my_dir.h>

#include <mysql/plugin.h>
#include <mysql.h>

/* Global variables */
static char* row_format= NULL;

/* prototypes */
static handler* memcache_create_handler(handlerton *hton,
                                        TABLE_SHARE *table, 
                                        MEM_ROOT *mem_root);
static int init_func(void *p);
bool memcache_show_status(handlerton *hton, 
                          THD* thd, 
                          stat_print_fn *stat_print,
                          enum ha_stat_type stat_type);
bool schema_table_store_record(THD *thd, TABLE *table);

int memcache_discover(handlerton *hton, THD* thd, 
                      const char *db, 
                      const char *name,
                      unsigned char** frmblob, 
                      size_t* frmlen);

/* Variables for memcache share methods */
static HASH memcache_open_tables; // Hash used to track open tables
pthread_mutex_t memcache_mutex;   // This is the mutex we use to init the hash

/* Variables for memcache server table */
static HASH memcache_server_hash; // Hash used to track open tables
pthread_mutex_t memcache_server_mutex;   // This is the mutex we use to init the hash

/*
  Function we use in the creation of our hash to get key.
*/
static unsigned char* memcache_get_key(MEMCACHE_SHARE *share, uint *length,
                             my_bool not_used __attribute__((unused)))
{
  *length=share->table_name_length;
  return (unsigned char*) share->table_name;
}

static unsigned char* memcache_server_get_key(memcache_server_st *server, uint *length,
                                     my_bool not_used __attribute__((unused)))
{
  *length= server->server_name_length;
  return (unsigned char*) server->server_name;
}

static int init_func(void *p)
{
  DBUG_ENTER("init_func");

  handlerton *memcache_hton= (handlerton *)p;
  xml_row_startup();

  VOID(pthread_mutex_init(&memcache_mutex,MY_MUTEX_INIT_FAST));
  (void) hash_init(&memcache_open_tables,system_charset_info,32,0,0,
                   (hash_get_key) memcache_get_key,0,0);

  VOID(pthread_mutex_init(&memcache_server_mutex,MY_MUTEX_INIT_FAST));
  (void) hash_init(&memcache_server_hash,system_charset_info,32,0,0,
                   (hash_get_key) memcache_server_get_key,0,0);

  memcache_hton->state= SHOW_OPTION_YES;
  memcache_hton->create= memcache_create_handler;
  memcache_hton->discover= memcache_discover;
  memcache_hton->flags= HTON_CAN_RECREATE;

  DBUG_RETURN(0);
}

static int done_func(void *p)
{
  int error= 0;
  DBUG_ENTER("done_func");

  if (memcache_open_tables.records)
    error= 1;
  hash_free(&memcache_open_tables);
  pthread_mutex_destroy(&memcache_mutex);

  hash_free(&memcache_server_hash);
  pthread_mutex_destroy(&memcache_server_mutex);

  xml_row_shutdown();

  DBUG_RETURN(0);
}

/*
  Example of simple lock controls. The "share" it creates is structure we will
  pass to each memcache handler. Do you have to have one of these? Well, you have
  pieces that are used for locking, and they are needed to function.
*/
static MEMCACHE_SHARE *get_share(const char *table_name, TABLE *table)
{
  MEMCACHE_SHARE *share;
  uint length;
  char *tmp_name;

  pthread_mutex_lock(&memcache_mutex);
  length=(uint) strlen(table_name);

  if (!(share=(MEMCACHE_SHARE*) hash_search(&memcache_open_tables,
                                           (unsigned char*) table_name,
                                           length)))
  {
    if (!(share=(MEMCACHE_SHARE *)
          my_multi_malloc(MYF(MY_WME | MY_ZEROFILL),
                          &share, sizeof(*share),
                          &tmp_name, length+1,
                          NullS)))
    {
      pthread_mutex_unlock(&memcache_mutex);
      return NULL;
    }

    share->use_count=0;
    if (!memcmp("NATIVE", table->s->comment.str, sizeof("NATIVE")))
      share->native= TRUE;
    share->use_count=0;
    share->table_name_length=length;
    share->table_name=tmp_name;
    strmov(share->table_name,table_name);
    if (my_hash_insert(&memcache_open_tables, (unsigned char*) share))
      goto error;
    thr_lock_init(&share->lock);
    pthread_mutex_init(&share->mutex,MY_MUTEX_INIT_FAST);
  }
  share->use_count++;
  pthread_mutex_unlock(&memcache_mutex);

  return share;

error:
  pthread_mutex_destroy(&share->mutex);
  my_free((unsigned char *) share, MYF(0));

  return NULL;
}


/*
  Free lock controls. We call this whenever we close a table. If the table had
  the last reference to the share then we free memory associated with it.
*/
static int free_share(MEMCACHE_SHARE *share)
{
  pthread_mutex_lock(&memcache_mutex);
  if (!--share->use_count)
  {
    hash_delete(&memcache_open_tables, (unsigned char*) share);
    thr_lock_delete(&share->lock);
    pthread_mutex_destroy(&share->mutex);
    my_free((unsigned char *) share, MYF(0));
  }
  pthread_mutex_unlock(&memcache_mutex);

  return 0;
}

memcache_server_st *get_server(char *server_name, uint server_name_length)
{
  memcache_server_st *server;

  pthread_mutex_lock(&memcache_server_mutex);

  if (!(server=(memcache_server_st *) hash_search(&memcache_server_hash,
                                           (unsigned char*) server_name,
                                           server_name_length)))
  {
  server= (memcache_server_st *)
    my_malloc(sizeof(memcache_server_st), MYF(MY_WME | MY_ZEROFILL));

    server->server_name= my_strndup(server_name, server_name_length,
                                    MYF(0));
    server->server_name_length= server_name_length;

    if (my_hash_insert(&memcache_server_hash, (unsigned char*) server))
      goto error;
  }
  server->use_count++;
  pthread_mutex_unlock(&memcache_server_mutex);

  return server;

error:
  my_free((unsigned char *) server, MYF(0));

  return NULL;
}


static int free_server(memcache_server_st *server)
{
  pthread_mutex_lock(&memcache_server_mutex);
  if (!--server->use_count)
  {
    hash_delete(&memcache_server_hash, (unsigned char*) server);
    my_free((unsigned char *) server, MYF(0));
  }
  pthread_mutex_unlock(&memcache_server_mutex);

  return 0;
}


static handler* memcache_create_handler(handlerton *hton,
                                        TABLE_SHARE *table, 
                                        MEM_ROOT *mem_root)
{
  return new (mem_root) ha_memcache(hton, table);
}


ha_memcache::ha_memcache(handlerton *hton, 
                         TABLE_SHARE *table_arg)
  :handler(hton, table_arg)
{
  rec_buff= NULL;
  alloced_rec_buff_length= 0;
}

/*
  If frm_error() is called then we will use this to to find out what file extentions
  exist for the storage engine. This is also used by the default rename_table and
  delete_table method in handler.cc.
*/
static const char *ha_memcache_exts[] = {
  NullS
};

const char **ha_memcache::bas_ext() const
{
  return ha_memcache_exts;
}


/*
  Used for opening tables. The name will be the name of the file.
  A table is opened when it needs to be opened. For instance
  when a request comes in for a select on the table (tables are not
  open and closed for each request, they are cached).

  Called from handler.cc by handler::ha_open(). The server opens all tables by
  calling ha_open() which then calls the handler specific open().
*/
int ha_memcache::open(const char *name, int mode, uint test_if_locked)
{
  char buffer[1024];
  char *begin, *end, *ptr;

  DBUG_ENTER("ha_memcache::open");

  if (open_connection(table->s->connect_string.str,
                      table->s->connect_string.length))
    DBUG_RETURN(ER_CONNECT_TO_FOREIGN_DATA_SOURCE);

  if (!(share = get_share(name, table)))
    DBUG_RETURN(1);
  thr_lock_data_init(&share->lock,&lock,NULL);

  ref_length= find_primary_key_length() + 4;

  DBUG_RETURN(0);
}

void ha_memcache::close_connection(void)
{
  DBUG_ENTER("ha_memcache::close_connection");
  memcached_free(memc);
  DBUG_VOID_RETURN;
}


int ha_memcache::open_connection(char *connect_string, unsigned int length)
{
  DBUG_ENTER("ha_memcache::open_connection");
  if (!length)
    DBUG_RETURN(1);

  memc= memcached_create(NULL);

#ifdef NOTDONE
  FOREIGN_SERVER *server;
  if ((server= get_server_by_name(connect_string)))
  {
    ptr=begin= server->host;
    end= begin + strlen(server->host);
  }
  else
#endif
  {
    memcached_return rc;
    memcached_server_st *servers;


    printf("servers %s\n", connect_string);

    servers= memcached_servers_parse(connect_string);

    unsigned int x;
    for (x= 0; x < memcached_server_list_count(servers); x++)
    {
      printf("\t%s : %u\n", servers[x].hostname, servers[x].port);
    }
    printf("\n");
    rc= memcached_server_push(memc, servers);
    assert(rc == MEMCACHED_SUCCESS);

    memcached_server_list_free(servers);
  }

  DBUG_RETURN(0);
}

/*
  Key lookup
*/

ha_rows ha_memcache::records_in_range(uint inx, key_range *min_key,
                         key_range *max_key)
{ 
  DBUG_ENTER("ha_memcache::records_in_range");
  DBUG_RETURN((ha_rows)1); 
}

int ha_memcache::index_read(unsigned char *buf, const unsigned char *key,
                            uint key_len, enum ha_rkey_function find_flag)
{
  DBUG_ENTER("ha_memcache::index_read");
  unsigned int rc= 0;

  ha_statistic_increment(&SSV::ha_read_key_count);

  rc= find_row(buf, key, key_len);
  DBUG_RETURN(rc);
}

int ha_memcache::read_range_first(const key_range *start_key,
                     const key_range *end_key,
                     bool eq_range, bool sorted)
{ 
  DBUG_ENTER("ha_memcache::read_range_first");
  int ret;

  /* We are not fetching a row, we are checking for existence */
  ret= find_row(table->record[0], start_key->key,
                  start_key->length);
  DBUG_RETURN(ret); 
}

int ha_memcache::read_range_next()
{ 
  DBUG_ENTER("ha_memcache::read_range_next");
  DBUG_RETURN(HA_ERR_END_OF_FILE); 
}

#ifdef NOT_DONE
const COND* ha_memcache::cond_push(const COND *cond) 
{ 
  DBUG_ENTER("cond_push");
  DBUG_RETURN(0);
}
#endif


int ha_memcache::update_row(const unsigned char *old_data, unsigned char *new_data)
{
  memcached_return rc;
  char prev_key[2048];
  size_t prev_key_length;
  DBUG_ENTER("ha_memcache::update_row");

  Field *key_field= table->key_info[table->s->primary_key].key_part->field;

  my_bitmap_map *org_bitmap= dbug_tmp_use_all_columns(table, table->read_set);

  if (table->timestamp_field_type & TIMESTAMP_AUTO_SET_ON_UPDATE)
    table->timestamp_field->set_time();
  ha_statistic_increment(&SSV::ha_update_count);

  pack_row(new_data);
  /* 
    We need to save our old key to make sure that the primary key has not been 
    updated if it has we will need to delete first.
  */
  memcpy(prev_key, current_key, current_key_length);
  prev_key_length= current_key_length;
  make_key(key_field);
  
  if (current_key_length != prev_key_length ||
      memcmp(current_key, prev_key, current_key_length))
  {
    /* This can fail if a concurrent delete happens */
    rc= memcached_add(memc, current_key, current_key_length,
                      (char *)rec_buff, rec_buff_length, 0, 0);
    if (rc == MEMCACHED_SUCCESS)
      /* If we fail at this, it means someone has already updated the key */
      rc= memcached_delete(memc, prev_key, prev_key_length, (time_t)0);
  }
  else
  {
    /* This can fail if a concurrent delete happens */
    rc= memcached_replace(memc, current_key, current_key_length,
                          (char *)rec_buff, rec_buff_length, 0, 0);
  }

  dbug_tmp_restore_column_map(table->read_set, org_bitmap);

  /* 
    If rc is anything other then success then something went wrong,
    how to handle this needs to be thought out.
  */
  DBUG_RETURN((rc == MEMCACHED_SUCCESS) ? 0 : -1);
}


int ha_memcache::delete_row(const unsigned char *buf)
{
  memcached_return rc;
  DBUG_ENTER("ha_memcache::delete_row");

  Field *key_field= table->key_info[table->s->primary_key].key_part->field;

  make_key(key_field);

  rc= memcached_delete(memc, current_key, current_key_length, (time_t)0);

  DBUG_RETURN((rc == MEMCACHED_SUCCESS) ? 0 : -1);
}


/*
  Closes a table. We call the free_share() function to free any resources
  that we have allocated in the "shared" structure.

  Called from sql_base.cc, sql_select.cc, and table.cc.
  In sql_select.cc it is only used to close up temporary tables or during
  the process where a temporary table is converted over to being a
  myisam table.
  For sql_base.cc look at close_data_tables().
*/
int ha_memcache::close(void)
{
  DBUG_ENTER("ha_memcache::close");
  close_connection();
  DBUG_RETURN(free_share(share));
}


int ha_memcache::rnd_init(bool scan)
{
  DBUG_ENTER("ha_memcache::rnd_init");
  DBUG_RETURN(0);
}


int ha_memcache::rnd_end()
{
  DBUG_ENTER("ha_memcache::rnd_end");
  DBUG_RETURN(0);
}

/* Reallocate buffer if needed */

bool ha_memcache::fix_rec_buff(size_t length)
{
  DBUG_ENTER("ha_memcache::fix_rec_buff");
  if (! rec_buff || length > alloced_rec_buff_length)
  {
    unsigned char *newptr;
    if (!(newptr= (unsigned char*)my_realloc((unsigned char *) rec_buff, length,
				    MYF(MY_ALLOW_ZERO_PTR))))
      DBUG_RETURN(1); /* purecov: inspected */
    rec_buff= newptr;
    alloced_rec_buff_length= length;
  }
 DBUG_RETURN(0);
}


/* Calculate max length needed for row */

/* Memory optimization would be to calculate Field completely */
size_t ha_memcache::max_row_length(void)
{
  size_t length= table->s->reclength + table->s->fields*2;
  uint *ptr, *end;

  DBUG_ENTER("ha_memcache::max_row_length");

  for (ptr= table->s->blob_field, end=ptr + table->s->blob_fields ;
       ptr != end ;
       ptr++)
  {
    length += 2 + ((Field_blob*)table->field[*ptr])->get_length();
  }
  DBUG_RETURN(length);
}


size_t ha_memcache::pack_row(unsigned char *record)
{
  DBUG_ENTER("ha_memcache::pack_row");

  if (!row_format)
  {
    unsigned char *ptr;

    if (fix_rec_buff(max_row_length()))
      DBUG_RETURN(HA_ERR_OUT_OF_MEM); /* purecov: inspected */

    /* Copy null bits */
    memcpy(rec_buff, record, table->s->null_bytes);
    ptr= rec_buff + table->s->null_bytes;

    for (Field **field=table->field ; *field ; field++)
    {
      if (!((*field)->is_null()))
        ptr=(unsigned char*) (*field)->pack((unsigned char*) ptr,
                                            (unsigned char*) record + (*field)->offset(record));
    }

    rec_buff_length= (size_t)(ptr - rec_buff);
  }
  else
  {
    xml_row_st engine;
    xml_row_st myrow, *rowptr;
    char *string;

    char content_buffer[1024];
    String content(content_buffer, sizeof(content_buffer),
                   &my_charset_bin);
    content.length(0);

    xml_row_create(&myrow);

    for (Field **field=table->field ; *field ; field++)
    {
      (*field)->val_str(&content);
      if ((*field)->is_null())
      {
        xml_row_field_add(&myrow, (char *)(*field)->field_name, NULL, 0, XML_ENGINE_IS_NULL);
      }
      else
      {
        (*field)->val_str(&content);
        xml_row_field_add(&myrow, (char *)(*field)->field_name, 
                          content.c_ptr_safe(), content.length(), 
                          XML_ENGINE_NOT_NULL);
      }
    }

    string= xml_row_serialize(&myrow, &rec_buff_length);

    if (fix_rec_buff(rec_buff_length))
      DBUG_RETURN(HA_ERR_OUT_OF_MEM); /* purecov: inspected */

    memcpy(rec_buff, string, rec_buff_length);

    xmlFree(string);

    xml_row_close(&myrow);
  }

  DBUG_RETURN(rec_buff_length);
}


void ha_memcache::unpack_row(unsigned char *record, char *reciever, size_t reciever_length)
{
  /* Copy null bits */
  if (!row_format)
  {
    const unsigned char *ptr= (const unsigned char *)reciever;
    memcpy(record, ptr, table->s->null_bytes);
    ptr+= table->s->null_bytes;
    for (Field **field=table->field ; *field ; field++)
      if (!((*field)->is_null()))
      {
        ptr= (*field)->unpack((unsigned char *)record + 
                              (*field)->offset(table->record[0]), ptr);
      }
  }
  else
  {
    xml_row_return rc;
    xml_row_st myrow;
    xml_row_field_st *row_ptr;

    xml_row_create(&myrow);

    if ((rc == xml_row_set(&myrow, reciever, reciever_length)) == XML_ROW_FAILURE)
      return;

    Field **field=table->field;

    row_ptr= xml_row_read(&myrow);

    for (; row_ptr; row_ptr= row_ptr->next) 
    {
      (*field)->store((char *)row_ptr->str, row_ptr->length, &my_charset_bin);
      (*field)->set_notnull();
      field++;
    }

    xml_row_close(&myrow);
  }
}


int ha_memcache::write_row(unsigned char *buf)
{
  memcached_return rc;
  DBUG_ENTER("ha_memcache::write_row");
  Field *key_field= table->key_info[table->s->primary_key].key_part->field;

  my_bitmap_map *org_bitmap= dbug_tmp_use_all_columns(table, table->read_set);

  ha_statistic_increment(&SSV::ha_write_count);

  if (table->timestamp_field_type & TIMESTAMP_AUTO_SET_ON_INSERT)
    table->timestamp_field->set_time();

  pack_row(buf);
  make_key(key_field);

  rc= memcached_add(memc, current_key, current_key_length,
             (char *)rec_buff, rec_buff_length,
             (time_t)0, (uint16_t)0);

  dbug_tmp_restore_column_map(table->read_set, org_bitmap);

  switch (rc)
  {
  case MEMCACHED_NOTSTORED:
    DBUG_RETURN(HA_ERR_FOUND_DUPP_KEY);
  case MEMCACHED_SUCCESS:
    DBUG_RETURN(0);
  default:
    DBUG_RETURN(-1);
  }
}


int ha_memcache::rnd_next(unsigned char *buf)
{
  DBUG_ENTER("ha_memcache::rnd_next");
  DBUG_RETURN(HA_ERR_END_OF_FILE);
}


void ha_memcache::position(const unsigned char *record)
{
  DBUG_ENTER("ha_memcache::position");
  unsigned int key_length;
  Field *key_field= table->key_info[table->s->primary_key].key_part->field;

  key_length= key_field->data_length();
  memcpy(ref, &key_length, sizeof(unsigned int));
  memcpy(ref+sizeof(unsigned int), key_field->ptr, key_length);

  DBUG_VOID_RETURN;
}


void ha_memcache::make_key(Field *primary)
{
  char *end_ptr= key_built_buffer;

  char attribute_buffer[1024];
  String attribute(attribute_buffer, sizeof(attribute_buffer),
                   &my_charset_bin);
  DBUG_ENTER("ha_memcache::make_key");

  current_key= key_built_buffer;

  primary->val_str(&attribute, &attribute);


  if (!share->native)
  {
    memcpy(key_built_buffer, table->s->db.str, table->s->db.length);
    end_ptr+= table->s->db.length;
    memcpy(end_ptr, table->s->table_name.str, table->s->table_name.length);
    end_ptr+= table->s->table_name.length;
  }

  memcpy(end_ptr, attribute.ptr(), attribute.length());
  end_ptr+= attribute.length();

  current_key_length= (size_t)(end_ptr - key_built_buffer);

  DBUG_VOID_RETURN;
}

int ha_memcache::find_row(unsigned char *buf, const unsigned char *key, uint key_len)
{
  DBUG_ENTER("ha_memcache::find_row");
  size_t ret_length= 0;
  char *ret;
  uint16_t flags;
  memcached_return rc;
  Field *key_field= table->key_info[table->s->primary_key].key_part->field;

  my_bitmap_map *old_map= dbug_tmp_use_all_columns(table, table->read_set);

  my_bitmap_map *old_write_map= dbug_tmp_use_all_columns(table, table->write_set);
  key_field->set_key_image(key, key_len);
  key_field->set_notnull();
  dbug_tmp_restore_column_map(table->write_set, old_write_map);

  /* Now we fetch the key we will use */
  make_key(key_field);

  ret= (char *)memcached_get(memc, (char *)current_key, 
                             current_key_length, &ret_length,
                             &flags, &rc);

  if (rc == MEMCACHED_SUCCESS)
  {
    unpack_row(buf, ret, ret_length);
    free(ret);
  }

  dbug_tmp_restore_column_map(table->read_set, old_map);

  DBUG_RETURN((rc == MEMCACHED_SUCCESS) ? 0 : HA_ERR_END_OF_FILE);
}

/* Used for ::open */
unsigned int ha_memcache::find_primary_key_length(void)
{
  DBUG_ENTER("ha_memcache::find_primary_key_length");
  unsigned int key_length;
  Field *key_field= table->key_info[table->s->primary_key].key_part->field;

  key_length= key_field->max_packed_col_length(key_field->field_length);
  if (!share->native)
  {
    key_length+= table->s->db.length;
    key_length+= table->s->table_name.length;
  }

  DBUG_RETURN(key_length);
}


/*
  This is like rnd_next, but you are given a position to use
  to determine the row. The position will be of the type that you stored in
  ref. You can use ha_get_ptr(pos,ref_length) to retrieve whatever key
  or position you saved when position() was called.
  Called from filesort.cc records.cc sql_insert.cc sql_select.cc sql_update.cc.
*/
int ha_memcache::rnd_pos(unsigned char * buf, unsigned char *pos)
{
  int ret;
  unsigned int key_length;
  DBUG_ENTER("ha_memcache::rnd_pos");

  memcpy(&key_length, pos, sizeof(unsigned int));
  ret= find_row(buf, pos + sizeof(unsigned int), key_length);
  DBUG_RETURN(ret); 
}


/*

*/
int ha_memcache::info(uint flag)
{
  DBUG_ENTER("ha_memcache::info");
  /* Without setting this to something high we won't get IN() to work */
  stats.records= 300000;
  DBUG_RETURN(0);
}

/*
*/

THR_LOCK_DATA **ha_memcache::store_lock(THD *thd,
                                       THR_LOCK_DATA **to,
                                       enum thr_lock_type lock_type)
{
  if (lock_type != TL_IGNORE && lock.type == TL_UNLOCK) 
  {

    /* If we are not doing a LOCK TABLE, then allow multiple
       writers */
    
    /* Since MEMCACHED does not currently have table locks
       this is treated as a ordinary lock */

    if ((lock_type >= TL_WRITE_CONCURRENT_INSERT &&
         lock_type <= TL_WRITE) && !(thd_in_lock_tables(thd)))
      lock_type= TL_WRITE_ALLOW_WRITE;
    
    /* In queries of type INSERT INTO t1 SELECT ... FROM t2 ...
       MySQL would use the lock TL_READ_NO_INSERT on t2, and that
       would conflict with TL_WRITE_ALLOW_WRITE, blocking all inserts
       to t2. Convert the lock to a normal read lock to allow
       concurrent inserts to t2. */
    
    if (lock_type == TL_READ_NO_INSERT && !(thd_in_lock_tables(thd)))
      lock_type= TL_READ;
    
    lock.type=lock_type;
  }
  *to++= &lock;
  return to;
}


/*
  create() is called to create a database. The variable name will have the name
  of the table. When create() is called you do not need to worry about opening
  the table. Also, the FRM file will have already been created so adjusting
  create_info will not do you any good. You can overwrite the frm file at this
  point if you wish to change the table definition, but there are no methods
  currently provided for doing that.

  Called from handle.cc by ha_create_table().
*/
int ha_memcache::create(const char *name, TABLE *table_arg,
                       HA_CREATE_INFO *create_info)
{
  unsigned int count;
  DBUG_ENTER("ha_memcache::create");

  File frm_file;                   /* File handler for readers */
  MY_STAT file_stat;  // Stat information for the data file
  unsigned char *frm_ptr;
  char name_buff[FN_REFLEN];

  fn_format(name_buff, name, "", ".frm",
            MY_REPLACE_EXT | MY_UNPACK_FILENAME);

  if (open_connection(create_info->connect_string.str,
                      create_info->connect_string.length))
    DBUG_RETURN(ER_CONNECT_TO_FOREIGN_DATA_SOURCE);

  /*
    Here is where we open up the frm and pass it to memcache to store 
  */
  if ((frm_file= my_open(name_buff, O_RDONLY, MYF(0))) > 0)
  {
    if (!my_fstat(frm_file, &file_stat, MYF(MY_WME)))
    {
      frm_ptr= (unsigned char *)my_malloc(sizeof(unsigned char) * file_stat.st_size , MYF(0));
      if (frm_ptr)
      {
        memcached_return rc;

        my_read(frm_file, frm_ptr, file_stat.st_size, MYF(0));
        // Save it for discovery
        rc= memcached_set(memc, (char *)table_arg->s->table_name.str, 
                          table_arg->s->table_name.length,
                          (char *)frm_ptr, file_stat.st_size, 0, 0);
        my_free((unsigned char *)frm_ptr, MYF(0));
      }
      /* Warning should go here that discovery will not work */
    }
    my_close(frm_file, MYF(0));
  }

  close_connection();

  DBUG_RETURN(0);
}


int memcache_discover(handlerton *hton, THD* thd, 
                      const char *db, 
                      const char *name,
                      unsigned char** frmblob, 
                      size_t* frmlen)
{
  DBUG_ENTER("memcache_discover");
  DBUG_PRINT("memcache_discover", ("db: %s, name: %s", db, name)); 
  size_t ret_length= 0;
  char *frm_ptr;
  char *ret;

  pthread_mutex_lock(&memcache_server_mutex);
  for (uint x= 0; x < memcache_server_hash.records; ++x)
  {
    memcached_return rc;
    memcached_st *dmc;
    uint16_t flags;

    memcache_server_st *server= 
      (memcache_server_st *) hash_element(&memcache_server_hash, x);

    dmc= memcached_create(NULL);

    rc= memcached_server_add(dmc, server->server_name, 0);

    ret= memcached_get(dmc, (char *)name, 
                       strlen(name), &ret_length,
                       &flags, &rc);
    if (ret_length)
    {
      *frmblob= (unsigned char *)my_strndup((const char *)ret, ret_length, MYF(0));
      free(ret);
      *frmlen= ret_length;
      break;
    }

    memcached_free(dmc);
  }
  pthread_mutex_unlock(&memcache_server_mutex);

  if (ret_length)
    DBUG_RETURN(0);

  my_errno= 0;
  DBUG_RETURN(1);
}

static int fill_memcache_server_schema(THD *thd, TABLE_LIST *tables, COND *cond)
{
  CHARSET_INFO *scs= system_charset_info;
  TABLE *table= (TABLE *) tables->table;
  memcached_st *memc_local;
  unsigned int x;
  DBUG_ENTER("fill_memcache_server_schema");
  int returnable= 0;

  pthread_mutex_lock(&memcache_server_mutex);
  for (x= 0; x < memcache_server_hash.records; ++x)
  {
    memcached_return rc;
    memcached_st *memc_local;
    memcached_stat_st *stats;
    memcache_server_st *server= 
      (memcache_server_st *) hash_element(&memcache_server_hash, x);

    memc_local= memcached_create(NULL);
    rc= memcached_server_add(memc_local, server->server_name, 0);
    stats= memcached_stat(memc_local, NULL, &rc);

    table->field[0]->store(server->server_name, 
                           server->server_name_length, 
                           scs);
    table->field[1]->store(server->use_count);
    table->field[2]->store(stats->curr_items);
    table->field[3]->store(stats->total_items);
//    table->field[4]->store(stats->u_chars);
    table->field[5]->store(stats->curr_connections);
    table->field[6]->store(stats->total_connections);
    table->field[7]->store(stats->connection_structures);
    table->field[8]->store(stats->cmd_get);
    table->field[9]->store(stats->cmd_set);
    table->field[10]->store(stats->get_hits);
    table->field[11]->store(stats->get_misses);
//    table->field[12]->store(stats->uchars_read);
//    table->field[13]->store(stats->uchars_written);
//    table->field[14]->store(stats->limit_maxunsigned chars);

    memcached_stat_free(memc_local, stats);
    memcached_free(memc_local);

    if (schema_table_store_record(thd, table))
    {
      returnable= 1;
      break;
    }
  }
  pthread_mutex_unlock(&memcache_server_mutex);

  DBUG_RETURN(returnable);
}

ST_FIELD_INFO memcache_server_field_info[]=
{
  {"NAME", 120, MYSQL_TYPE_STRING, 0, 0, "Name"},
  {"COUNT", 4, MYSQL_TYPE_LONG, 0, 0, "Count"},
  {"CURRENT_ITEMS", 4, MYSQL_TYPE_LONG, 0, 0, "Current Items"},
  {"TOTAL_ITEMS", 4, MYSQL_TYPE_LONG, 0, 0, "Total Items"},
  {"BYTES", 4, MYSQL_TYPE_LONG, 0, 0, "Bytes"},
  {"CURRENT_CONNECTIONS", 4, MYSQL_TYPE_LONG, 0, 0, "Current Connections"},
  {"TOTAL_CONNECTIONS", 4, MYSQL_TYPE_LONG, 0, 0, "Total Connections"},
  {"CONNECTION_STRUCTURES", 4, MYSQL_TYPE_LONG, 0, 0, "Connection Structure"},
  {"GETS", 4, MYSQL_TYPE_LONG, 0, 0, "Gets"},
  {"SETS", 4, MYSQL_TYPE_LONG, 0, 0, "Sets"},
  {"HITS", 4, MYSQL_TYPE_LONG, 0, 0, "Hits"},
  {"MISSES", 4, MYSQL_TYPE_LONG, 0, 0, "Misses"},
  {"BYTES_READ", 4, MYSQL_TYPE_LONG, 0, 0, "Bytes Read"},
  {"BYTES_WRITTEN", 4, MYSQL_TYPE_LONG, 0, 0, "Bytes Written"},
  {"LIMIT_MAXBYTES", 4, MYSQL_TYPE_LONG, 0, 0, "Limit Maxunsigned chars"},
  {0, 0, MYSQL_TYPE_STRING, 0, 0, 0}
};

static int is_init(void *p)
{
  DBUG_ENTER("is_init");
  ST_SCHEMA_TABLE *schema= (ST_SCHEMA_TABLE *)p;

  struct st_plugin_int *plugin= (struct st_plugin_int *)p;

  schema->fields_info= memcache_server_field_info;
  schema->fill_table= fill_memcache_server_schema;

  DBUG_RETURN(0);
}


static int is_deinit(void *p)
{
  DBUG_ENTER("is_deinit");
  DBUG_RETURN(0);
}

static MYSQL_SYSVAR_STR(row_format, row_format,
  PLUGIN_VAR_RQCMDARG | PLUGIN_VAR_READONLY,
  "Row format to use, native(default) or xml.", NULL, NULL, NULL);

static struct st_mysql_sys_var* system_variables[]= {
  MYSQL_SYSVAR(row_format),
  NULL
};

struct st_mysql_storage_engine memcache_storage_engine=
{ MYSQL_HANDLERTON_INTERFACE_VERSION };

struct st_mysql_information_schema memcache_server_is=
{ MYSQL_INFORMATION_SCHEMA_INTERFACE_VERSION };


mysql_declare_plugin(memcache)
{
  MYSQL_STORAGE_ENGINE_PLUGIN,
  &memcache_storage_engine,
  "MEMCACHE",
  "Brian Aker, Tangent Org",
  "Simple Interface for working with memcache as a storage engine",
  PLUGIN_LICENSE_GPL,
  init_func, /* Plugin Init */
  done_func, /* Plugin Deinit */
  0x0006,
  NULL,                       /* status variables                */
  NULL,                       /* system variables                */
  NULL                        /* config options                  */
},
{
  MYSQL_INFORMATION_SCHEMA_PLUGIN,
  &memcache_server_is,
  "memcache_servers",
  "Brian Aker",
  "Active Memached Servers.",
  PLUGIN_LICENSE_GPL,
  is_init, /* Plugin Init */
  is_deinit, /* Plugin Deinit */
  0x0007,
  NULL,                       /* status variables                */
  system_variables, /* system variables */
  NULL                        /* config options                  */
}
mysql_declare_plugin_end;

