/* Copyright (C) 2003 MySQL AB

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/*
  Please read ha_exmple.cc before reading this file.
  Please keep in mind that the memcache storage engine implements all methods
  that are required to be implemented. handler.h has a full list of methods
  that you can implement.
*/

#ifdef USE_PRAGMA_INTERFACE
#pragma interface			/* gcc class implementation */
#endif

#define HAVE_DEBUG 1
#include <libmemcached/memcached.h>
#include "xmlrow.h"

/*
  MEMCACHE_SHARE is a structure that will be shared amoung all open handlers
  The memcache implements the minimum of what you will probably need.
*/
typedef struct st_memcache_share {
  char *table_name;
  uint table_name_length, use_count;
  pthread_mutex_t mutex;
  bool native;
  THR_LOCK lock;
} MEMCACHE_SHARE;

typedef struct memcache_server_st {
  char *server_name;
  uint server_name_length, use_count;
} memcache_server_st;

/*
  Class definition for the storage engine
*/
class ha_memcache: public handler
{
  THR_LOCK_DATA lock;      /* MySQL lock */
  MEMCACHE_SHARE *share;    /* Shared lock info */
  memcached_st *memc;
  char key_built_buffer[2048];
  size_t current_key_length;
  char *current_key;
  unsigned char *rec_buff;
  size_t alloced_rec_buff_length;
  size_t rec_buff_length;

  int open_connection(char *connect_string, unsigned int length);
  void close_connection(void);
  size_t pack_row(unsigned char *record);
  void unpack_row(unsigned char *record, char *reciever, size_t reciever_length);
  bool fix_rec_buff(size_t length);
  size_t max_row_length();
  unsigned int find_primary_key_length(void);
  int find_row(unsigned char *buf, const unsigned char *key, uint key_len);
  Field **find_primary_key(void);
  void make_key(Field *primary);

public:
  ha_memcache(handlerton *hton, TABLE_SHARE *table_arg);
  ~ha_memcache()
  {
  }
  /* The name that will be used for display purposes */
  const char *table_type() const { return "MEMCACHE"; }
  /*
    The name of the index type that will be used for display
    don't implement this method unless you really have indexes
   */
  const char **bas_ext() const;
  /*
    This is a list of flags that says what the storage engine
    implements. The current table flags are documented in
    handler.h
  */
  ulonglong table_flags() const
  {
    return (HA_REQUIRE_PRIMARY_KEY|HA_BINLOG_ROW_CAPABLE|HA_NO_TRANSACTIONS);
  }


  /*
    This is a bitmap of flags that says how the storage engine
    implements indexes. The current index flags are documented in
    handler.h. If you do not implement indexes, just return zero
    here.

    part is the key part to check. First key part is 0
    If all_parts it's set, MySQL want to know the flags for the combined
    index up to and including 'part'.
  */
  ulong index_flags(uint inx, uint part, bool all_parts) const
  {
    return HA_ONLY_WHOLE_INDEX;
  }


  /*
    What we return when asked about index type
  */
  const char *index_type(uint inx)
  {
    return ("HASH");
  }

  /*
    unireg.cc will call the following to make sure that the storage engine can
    handle the data it is about to send.

    Return *real* limits of your storage engine here. MySQL will do
    min(your_limits, MySQL_limits) automatically

    There is no need to implement ..._key_... methods if you don't suport
    indexes.
  */
  uint max_supported_record_length() const { return HA_MAX_REC_LENGTH; }
  uint max_supported_keys()          const { return 1; }
  uint max_supported_key_parts()     const { return 1; }
  uint max_supported_key_length()    const { return 256; } //memcached is maxed out at 255
  /*
    Everything below are methods that we implment in ha_memcache.cc.

    Most of these methods are not obligatory, skip them and
    MySQL will treat them as not implemented
  */
  int open(const char *name, int mode, uint test_if_locked);    // required
  int close(void);                                              // required

  /*
    unlike index_init(), rnd_init() can be called two times
    without rnd_end() in between (it only makes sense if scan=1).
    then the second call should prepare for the new table scan
    (e.g if rnd_init allocates the cursor, second call should
    position it to the start of the table, no need to deallocate
    and allocate it again
  */
  int write_row(unsigned char * buf);
  int rnd_init(bool scan);                                      //required
  int rnd_end();
  int rnd_next(unsigned char *buf);                                      //required
  int rnd_pos(unsigned char * buf, unsigned char *pos);                           //required
  int delete_row(const unsigned char *buf);
  int update_row(const unsigned char *old_data, unsigned char *new_data);
  int index_read(unsigned char * buf, const unsigned char * key,
                 uint key_len, enum ha_rkey_function find_flag);
  int index_next(unsigned char * buf)
  { return  HA_ERR_END_OF_FILE; }
  ha_rows records_in_range(uint inx, key_range *min_key,
                           key_range *max_key);

  int read_range_first(const key_range *start_key,
                       const key_range *end_key,
                       bool eq_range, bool sorted);
  int read_range_next();
#ifdef NOT_DONE
  const COND* ha_memcache::cond_push(const COND *cond);
#endif

  void position(const unsigned char *record);                            //required
  int info(uint);                                              //required

  int create(const char *name, TABLE *form,
             HA_CREATE_INFO *create_info);                      //required

  THR_LOCK_DATA **store_lock(THD *thd, THR_LOCK_DATA **to,
                             enum thr_lock_type lock_type);     //required
};

